/*
 * device driver for philips saa7134 based TV cards
 * video4linux video interface
 *
 * (c) 2001,02 Gerd Knorr <kraxel@bytesex.org> [SuSE Labs]
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define __NO_VERSION__ 1

#include <linux/init.h>
#include <linux/list.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/slab.h>

#include "saa7134-reg.h"
#include "saa7134.h"

/* ------------------------------------------------------------------ */

static unsigned int vbi_debug  = 0;
MODULE_PARM(vbi_debug,"i");
MODULE_PARM_DESC(vbi_debug,"enable debug messages [vbi]");

static unsigned int vbibufs = 4;
MODULE_PARM(vbibufs,"i");
MODULE_PARM_DESC(vbibufs,"number of vbi buffers, range 2-32");

#define dprintk(fmt, arg...)	if (vbi_debug) \
	printk(KERN_DEBUG "%s/vbi: " fmt, dev->name, ## arg)

/* ------------------------------------------------------------------ */

#define VBI_LINE_COUNT     16
#define VBI_LINE_LENGTH  2048
#define VBI_SCALE       0x200

static void task_init(struct saa7134_dev *dev, struct saa7134_buf *buf,
		      int task)
{
	struct saa7134_tvnorm *norm = dev->tvnorm;
	
	/* setup video scaler */
	saa_writeb(SAA7134_VBI_H_START1(task), norm->h_start     &  0xff);
	saa_writeb(SAA7134_VBI_H_START2(task), norm->h_start     >> 8);
	saa_writeb(SAA7134_VBI_H_STOP1(task),  norm->h_stop      &  0xff);
	saa_writeb(SAA7134_VBI_H_STOP2(task),  norm->h_stop      >> 8);
	saa_writeb(SAA7134_VBI_V_START1(task), norm->vbi_v_start &  0xff);
	saa_writeb(SAA7134_VBI_V_START2(task), norm->vbi_v_start >> 8);
	saa_writeb(SAA7134_VBI_V_STOP1(task),  norm->vbi_v_stop  &  0xff);
	saa_writeb(SAA7134_VBI_V_STOP2(task),  norm->vbi_v_stop  >> 8);

	saa_writeb(SAA7134_VBI_H_SCALE_INC1(task),        VBI_SCALE & 0xff);
	saa_writeb(SAA7134_VBI_H_SCALE_INC2(task),        VBI_SCALE >> 8);
	saa_writeb(SAA7134_VBI_PHASE_OFFSET_LUMA(task),   0x00);
	saa_writeb(SAA7134_VBI_PHASE_OFFSET_CHROMA(task), 0x00);

	saa_writeb(SAA7134_VBI_H_LEN1(task), buf->vb.width   & 0xff);
	saa_writeb(SAA7134_VBI_H_LEN2(task), buf->vb.width   >> 8);
	saa_writeb(SAA7134_VBI_V_LEN1(task), buf->vb.height  & 0xff);
	saa_writeb(SAA7134_VBI_V_LEN2(task), buf->vb.height  >> 8);

	saa_andorb(SAA7134_DATA_PATH(task), 0xc0, 0x00);
}

/* ------------------------------------------------------------------ */

static int buffer_activate(struct saa7134_dev *dev, struct saa7134_buf *buf)
{
	unsigned long control;

	dprintk("buffer_activate [%p]\n",buf);
	buf->vb.state = STATE_ACTIVE;

	task_init(dev,buf,TASK_A);
	task_init(dev,buf,TASK_B);
	saa_writeb(SAA7134_OFMT_DATA_A, 0x06);
	saa_writeb(SAA7134_OFMT_DATA_B, 0x06);

	/* DMA: setup channel 2+3 (= VBI Task A+B) */
	control = SAA7134_RS_CONTROL_BURST_16 |
		SAA7134_RS_CONTROL_ME |
		(buf->pt.dma >> 12);
	saa_writel(SAA7134_RS_BA1(2),buf->pt.offset);
	saa_writel(SAA7134_RS_BA2(2),buf->pt.offset + buf->vb.size/2);
	saa_writel(SAA7134_RS_PITCH(2),buf->vb.width);
	saa_writel(SAA7134_RS_CONTROL(2),control);
	saa_writel(SAA7134_RS_BA1(3),buf->pt.offset);
	saa_writel(SAA7134_RS_BA2(3),buf->pt.offset + buf->vb.size/2);
	saa_writel(SAA7134_RS_PITCH(3),buf->vb.width);
	saa_writel(SAA7134_RS_CONTROL(3),control);

	/* start DMA */
	saa7134_set_dmabits(dev);
	mod_timer(&dev->vbi_q.timeout, jiffies+BUFFER_TIMEOUT);

	return 0;
}

static int buffer_prepare(struct file *file, struct videobuf_buffer *vb, int field)
{
	struct saa7134_dev *dev = file->private_data;
	struct saa7134_buf *buf = (struct saa7134_buf *)vb;
	struct saa7134_tvnorm *norm = dev->tvnorm;
	int lines, llength, size, err;

	lines   = norm->vbi_v_stop - norm->vbi_v_start +1;
	if (lines > VBI_LINE_COUNT)
		lines = VBI_LINE_COUNT;
#if 1
	llength = VBI_LINE_LENGTH;
#else
	llength = (norm->h_stop - norm->h_start +1) * 2;
	if (llength > VBI_LINE_LENGTH)
		llength = VBI_LINE_LENGTH;
#endif
	size = lines * llength * 2;
	if (0 != buf->vb.baddr  &&  buf->vb.bsize < size)
		return -EINVAL;

	if (buf->vb.size != size)
		saa7134_dma_free(dev,buf);

	if (STATE_NEEDS_INIT == buf->vb.state) {
		buf->vb.width  = llength;
		buf->vb.height = lines;
		buf->vb.size   = size;

		err = videobuf_iolock(dev->pci,&buf->vb);
		if (err)
			goto oops;
		err = saa7134_pgtable_build(dev->pci,&buf->pt,
					    buf->vb.dma.sglist,
					    buf->vb.dma.sglen);
		if (err)
			goto oops;
	}
	buf->vb.state = STATE_PREPARED;
	buf->odd_seen = 0;
	buf->activate = buffer_activate;
	buf->vb.field = VBUF_FIELD_ODD | VBUF_FIELD_EVEN;
	return 0;

 oops:
	saa7134_dma_free(dev,buf);
	return err;
}

static int
buffer_setup(struct file *file, int *count, int *size)
{
	struct saa7134_dev *dev = file->private_data;
	int llength,lines;
	
	lines   = dev->tvnorm->vbi_v_stop - dev->tvnorm->vbi_v_start +1;
#if 1
	llength = VBI_LINE_LENGTH;
#else
	llength = (norm->h_stop - norm->h_start +1) * 2;
	if (llength > VBI_LINE_LENGTH)
		llength = VBI_LINE_LENGTH;
#endif
	*size = lines * llength * 2;
	if (0 == *count)
		*count = vbibufs;
	return 0;
}

static void buffer_queue(struct file *file, struct videobuf_buffer *vb)
{
	struct saa7134_dev *dev = file->private_data;
	struct saa7134_buf *buf = (struct saa7134_buf *)vb;
	
	saa7134_buffer_queue(dev,&dev->vbi_q,buf);
}

static void buffer_release(struct file *file, struct videobuf_buffer *vb)
{
	struct saa7134_dev *dev = file->private_data;
	struct saa7134_buf *buf = (struct saa7134_buf *)vb;
	
	saa7134_dma_free(dev,buf);
}

static struct videobuf_queue_ops vbi_qops = {
	buf_setup:    buffer_setup,
	buf_prepare:  buffer_prepare,
	buf_queue:    buffer_queue,
	buf_release:  buffer_release,
};

/* ------------------------------------------------------------------ */

static int vbi_open(struct inode *inode, struct file *file)
{
	unsigned int minor = minor(inode->i_rdev);
	struct saa7134_dev *h,*dev = NULL;
	struct list_head *list;
	int err;

	list_for_each(list,&saa7134_devlist) {
		h = list_entry(list, struct saa7134_dev, devlist);
		if (h->vbi_dev.minor == minor)
			dev = h;
	}
	if (NULL == dev)
		return -ENODEV;

	dprintk("open #1 [minor=%d]\n",minor);
	down(&dev->vbi.q.lock);
	dprintk("open #2 [users=%d,havelock]\n",dev->vbi.users);
	err = -EBUSY;
	if (dev->vbi.users)
		goto done;

	dev->vbi.users++;
	file->private_data = dev;
	err = 0;

 done:
	up(&dev->vbi.q.lock);
	return err;
}

static int vbi_release(struct inode *inode, struct file *file)
{
	struct saa7134_dev *dev = file->private_data;

	if (dev->vbi.q.streaming)
		videobuf_streamoff(file,&dev->vbi.q);
	down(&dev->vbi.q.lock);
	if (dev->vbi.q.reading)
		videobuf_read_stop(file,&dev->vbi.q);
	dev->vbi.users--;
	dprintk("release [users=%d]\n",dev->vbi.users);
	up(&dev->vbi.q.lock);
	return 0;
}

static ssize_t
vbi_read(struct file *file, char *data, size_t count, loff_t *ppos)
{
	struct saa7134_dev *dev = file->private_data;

	dprintk("read [count=%d]\n",count);
	return videobuf_read_stream(file, &dev->vbi.q, data, count, ppos, 1);
}

static unsigned int
vbi_poll(struct file *file, struct poll_table_struct *wait)
{
	struct saa7134_dev *dev = file->private_data;

	return videobuf_poll_stream(file, &dev->vbi.q, wait);
}

static int
vbi_mmap(struct file *file, struct vm_area_struct * vma)
{
	struct saa7134_dev *dev = file->private_data;

	return videobuf_mmap_mapper(vma, &dev->vbi.q);
}

/*
 * This function is _not_ called directly, but from
 * video_generic_ioctl (and maybe others).  userspace
 * copying is done already, arg is a kernel pointer.
 */
static int vbi_do_ioctl(struct inode *inode, struct file *file,
			unsigned int cmd, void *arg)
{
	struct saa7134_dev *dev = file->private_data;

	if (vbi_debug > 1)
		saa7134_print_ioctl(dev->name,cmd);
	switch (cmd) {
	case VIDIOC_QUERYCAP:
	{
		struct v4l2_capability *cap = arg;

		memset(cap,0,sizeof(*cap));
                strcpy(cap->name, dev->name);
		cap->type = V4L2_TYPE_VBI;
		cap->flags = V4L2_FLAG_TUNER | V4L2_FLAG_READ |
			V4L2_FLAG_STREAMING | V4L2_FLAG_SELECT;
		return 0;
	}

	/* --- common ioctls ---------------------------------------- */
	case VIDIOC_ENUMSTD:
	case VIDIOC_G_STD:
	case VIDIOC_S_STD:
	case VIDIOC_ENUMINPUT:
	case VIDIOC_G_INPUT:
	case VIDIOC_S_INPUT:
	case VIDIOC_G_TUNER:
	case VIDIOC_S_TUNER:
	case VIDIOC_G_FREQ:
	case VIDIOC_S_FREQ:
		return saa7134_common_ioctl(dev,cmd,arg);
		
	/* --- capture ioctls ---------------------------------------- */
	case VIDIOC_G_FMT:
	{
		struct v4l2_format *f = arg;
		struct saa7134_tvnorm *norm = dev->tvnorm;

		memset(f,0,sizeof(*f));
		f->type = V4L2_BUF_TYPE_VBI;
		f->fmt.vbi.sampling_rate = 6750000 * 4;
		f->fmt.vbi.samples_per_line = VBI_LINE_LENGTH;
		f->fmt.vbi.sample_format = V4L2_VBI_SF_UBYTE;
		f->fmt.vbi.offset = 64 * 4;
		f->fmt.vbi.start[0] = norm->vbi_v_start;
		f->fmt.vbi.count[0] = norm->vbi_v_stop - norm->vbi_v_start +1;
		f->fmt.vbi.start[1] = norm->video_v_stop + norm->vbi_v_start + 1;
		f->fmt.vbi.count[1] = f->fmt.vbi.count[0];
		f->fmt.vbi.flags = 0; /* VBI_UNSYNC VBI_INTERLACED */;
		return 0;
	}
#if 0
	case VIDIOC_S_FMT:
	{
		struct v4l2_format *f = arg;

		if ((f->type & V4L2_BUF_TYPE_field) != V4L2_BUF_TYPE_VBI)
			return -EINVAL;
		/* TODO */
		return 0;
	}
#endif

	case VIDIOC_REQBUFS:
		return videobuf_reqbufs(file,&dev->vbi.q,arg);

	case VIDIOC_QUERYBUF:
		return videobuf_querybuf(&dev->vbi.q,arg);

	case VIDIOC_QBUF:
		return videobuf_qbuf(file,&dev->vbi.q,arg);

	case VIDIOC_DQBUF:
		return videobuf_dqbuf(file,&dev->vbi.q,arg);

	case VIDIOC_STREAMON:
		return videobuf_streamon(file,&dev->vbi.q);

	case VIDIOC_STREAMOFF:
		return videobuf_streamoff(file,&dev->vbi.q);

	default:
		return -ENOIOCTLCMD;
	}
	return 0;
}

static int vbi_ioctl(struct inode *inode, struct file *file,
		     unsigned int cmd, unsigned long arg)
{
	return video_usercopy(inode, file, cmd, arg, vbi_do_ioctl);
}

static struct file_operations vbi_fops =
{
	owner:	  THIS_MODULE,
	open:	  vbi_open,
	release:  vbi_release,
	read:	  vbi_read,
	poll:	  vbi_poll,
	mmap:	  vbi_mmap,
	ioctl:	  vbi_ioctl,
	llseek:   no_llseek,
};

/* ----------------------------------------------------------- */
/* exported stuff                                              */

struct video_device saa7134_vbi_template =
{
	name:          "saa7134-vbi",
	type:          VID_TYPE_TUNER|VID_TYPE_TELETEXT,
	type2:         V4L2_TYPE_VBI,
	hardware:      0,
	fops:          &vbi_fops,
	minor:         -1,
};

int saa7134_vbi_init(struct saa7134_dev *dev)
{
	INIT_LIST_HEAD(&dev->vbi_q.queue);
	dev->vbi_q.timeout.function = saa7134_buffer_timeout;
	dev->vbi_q.timeout.data     = (unsigned long)(&dev->vbi_q);
	dev->vbi_q.dev              = dev;
	videobuf_queue_init(&dev->vbi.q, &vbi_qops, dev->pci, &dev->slock,
			    V4L2_BUF_TYPE_VBI,
			    sizeof(struct saa7134_buf));
	
	if (vbibufs < 2)
		vbibufs = 2;
	if (vbibufs > VIDEO_MAX_FRAME)
		vbibufs = VIDEO_MAX_FRAME;
	return 0;
}

int saa7134_vbi_fini(struct saa7134_dev *dev)
{
	/* nothing */
	return 0;
}

void saa7134_irq_vbi_done(struct saa7134_dev *dev, unsigned long status)
{
	spin_lock(&dev->slock);
	if (dev->vbi_q.curr) {
		dev->vbi.fieldcount++;
		/* make sure we have seen both fields */
		if ((status & 0x10) == 0x10) {
			dev->vbi_q.curr->odd_seen = 1;
			goto done;
		}
		if (!dev->vbi_q.curr->odd_seen)
			goto done;

		dev->vbi_q.curr->vb.field_count = dev->vbi.fieldcount;
		saa7134_buffer_finish(dev,&dev->vbi_q,STATE_DONE);
	}
	saa7134_buffer_next(dev,&dev->vbi_q);

 done:
	spin_unlock(&dev->slock);
}

/* ----------------------------------------------------------- */
/*
 * Local variables:
 * c-basic-offset: 8
 * End:
 */
