/*
 * device driver for philips saa7134 based TV cards
 * video4linux video interface
 *
 * (c) 2001,02 Gerd Knorr <kraxel@bytesex.org> [SuSE Labs]
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define __NO_VERSION__ 1

#include <linux/init.h>
#include <linux/list.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/slab.h>

#include "saa7134-reg.h"
#include "saa7134.h"
#include "tuner.h"
#include "audiochip.h"

/* ------------------------------------------------------------------ */

static unsigned int video_debug  = 0;
static unsigned int gbuffers     = 8;
static unsigned int gbufsize     = 768*576*4;
static unsigned int gbufsize_max = 768*576*4;
MODULE_PARM(video_debug,"i");
MODULE_PARM_DESC(video_debug,"enable debug messages [video]");
MODULE_PARM(gbuffers,"i");
MODULE_PARM_DESC(gbuffers,"number of capture buffers, range 2-32");
MODULE_PARM(gbufsize,"i");
MODULE_PARM_DESC(gbufsize,"size of the capture buffers");

#define dprintk(fmt, arg...)	if (video_debug) \
	printk(KERN_DEBUG "%s/video: " fmt, dev->name, ## arg)

/* ------------------------------------------------------------------ */
/* data structs for video                                             */

static int video_out[][9] = {
	[CCIR656] = { 0x00, 0xb1, 0x00, 0xa1, 0x00, 0x04, 0x06, 0x00, 0x00 },
};
		
static struct saa7134_format formats[] = {
	{
		name:     "8 bpp gray",
		fourcc:   V4L2_PIX_FMT_GREY,
		depth:    8,
		pm:       0x06,
	},{
		name:     "15 bpp RGB, le",
		fourcc:   V4L2_PIX_FMT_RGB555,
		depth:    16,
		pm:       0x13 | 0x80,
	},{
		name:     "15 bpp RGB, be",
		fourcc:   V4L2_PIX_FMT_RGB555X,
		depth:    16,
		pm:       0x13 | 0x80,
		bswap:    1,
	},{
		name:     "16 bpp RGB, le",
		fourcc:   V4L2_PIX_FMT_RGB565,
		depth:    16,
		pm:       0x10 | 0x80,
	},{
		name:     "16 bpp RGB, be",
		fourcc:   V4L2_PIX_FMT_RGB565X,
		depth:    16,
		pm:       0x10 | 0x80,
		bswap:    1,
	},{
		name:     "24 bpp RGB, le",
		fourcc:   V4L2_PIX_FMT_BGR24,
		depth:    24,
		pm:       0x11,
	},{
		name:     "32 bpp RGB, le",
		fourcc:   V4L2_PIX_FMT_BGR32,
		depth:    32,
		pm:       0x12,
	},{
		name:     "32 bpp RGB, be",
		fourcc:   V4L2_PIX_FMT_RGB32,
		depth:    32,
		pm:       0x12,
		bswap:    1,
		wswap:    1,
	},{
		name:     "4:2:2 packed, YUYV",
		fourcc:   V4L2_PIX_FMT_YUYV,
		depth:    16,
		pm:       0x00,
		bswap:    1,
		yuv:      1,
	},{
		name:     "4:2:2 packed, UYVY",
		fourcc:   V4L2_PIX_FMT_UYVY,
		depth:    16,
		pm:       0x00,
		yuv:      1,
	},{
		name:     "4:2:2 planar, Y-Cb-Cr",
		fourcc:   V4L2_PIX_FMT_YUV422P,
		depth:    16,
		pm:       0x09,
		yuv:      1,
		planar:   1,
		hshift:   1,
		vshift:   0,
	},{
		name:     "4:2:0 planar, Y-Cb-Cr",
		fourcc:   V4L2_PIX_FMT_YUV420,
		depth:    12,
		pm:       0x0a,
		yuv:      1,
		planar:   1,
		hshift:   1,
		vshift:   1,
	}
};
#define FORMATS (sizeof(formats)/sizeof(struct saa7134_format))


static struct saa7134_tvnorm tvnorms[] = {
	{
		name:          "PAL-BGHI",
		id:            V4L2_STD_PAL,
		width:         720,
		height:        576,

		sync_control:  0x18,
		luma_control:  0x40,
		chroma_ctrl1:  0x81,
		chroma_gain:   0x2a,
		chroma_ctrl2:  0x06,

		h_start:       0,
		h_stop:        719,
		video_v_start: 24,
		video_v_stop:  311,
		vbi_v_start:   7-3,  /* FIXME */
		vbi_v_stop:    22-3,
	},{
		name:          "NTSC-M",
		id:            V4L2_STD_NTSC,
		width:         720,
		height:        480,

		sync_control:  0x59,
		luma_control:  0x40,
		chroma_ctrl1:  0x89,
		chroma_gain:   0x2a,
		chroma_ctrl2:  0x0e,

		h_start:       0,
		h_stop:        719,
		video_v_start: 22,
		video_v_stop:  22+240,
		vbi_v_start:   10, /* FIXME */
		vbi_v_stop:    21, /* FIXME */
	},{
		name:          "SECAM",
		id:            V4L2_STD_SECAM,
		width:         720,
		height:        576,

		sync_control:  0x58,
		luma_control:  0x1b,
		chroma_ctrl1:  0xd1,
		chroma_gain:   0x80,
		chroma_ctrl2:  0x00,

		h_start:       0,
		h_stop:        719,
		video_v_start: 24,
		video_v_stop:  311,
		vbi_v_start:   7,
		vbi_v_stop:    22,
	},{
		name:          "AUTO",
		id:            -1,
		width:         768,
		height:        576,

		sync_control:  0x98,
		luma_control:  0x40,
		chroma_ctrl1:  0x8b,
		chroma_gain:   0x00,
		chroma_ctrl2:  0x00,

		h_start:       0,
		h_stop:        719,
		video_v_start: 24,
		video_v_stop:  311,
		vbi_v_start:   7,
		vbi_v_stop:    22,
	}
};
#define TVNORMS (sizeof(tvnorms)/sizeof(struct saa7134_tvnorm))


#define V4L2_CID_PRIVATE_INVERT      (V4L2_CID_PRIVATE_BASE + 0)
#define V4L2_CID_PRIVATE_INTERLACE   (V4L2_CID_PRIVATE_BASE + 1)
#define V4L2_CID_PRIVATE_MIRROR      (V4L2_CID_PRIVATE_BASE + 2)
#define V4L2_CID_PRIVATE_Y_ODD       (V4L2_CID_PRIVATE_BASE + 3)
#define V4L2_CID_PRIVATE_Y_EVEN      (V4L2_CID_PRIVATE_BASE + 4)
#define V4L2_CID_PRIVATE_LASTP1      (V4L2_CID_PRIVATE_BASE + 5)

static const struct v4l2_queryctrl no_ctrl = {
	name:  "42",
	flags: V4L2_CTRL_FLAG_DISABLED,
};
static const struct v4l2_queryctrl video_ctrls[] = {
	/* --- video --- */
	{
		id:            V4L2_CID_BRIGHTNESS,
		name:          "Brightness",
		minimum:       0,
		maximum:       255,
		step:          1,
		default_value: 128,
		type:          V4L2_CTRL_TYPE_INTEGER,
		category:      V4L2_CTRL_CAT_VIDEO,
		group:         "Video",
	},{
		id:            V4L2_CID_CONTRAST,
		name:          "Contrast",
		minimum:       0,
		maximum:       127,
		step:          1,
		default_value: 68,
		type:          V4L2_CTRL_TYPE_INTEGER,
		category:      V4L2_CTRL_CAT_VIDEO,
		group:         "Video",
	},{
		id:            V4L2_CID_SATURATION,
		name:          "Saturation",
		minimum:       0,
		maximum:       127,
		step:          1,
		default_value: 64,
		type:          V4L2_CTRL_TYPE_INTEGER,
		category:      V4L2_CTRL_CAT_VIDEO,
		group:         "Video",
	},{
		id:            V4L2_CID_HUE,
		name:          "Hue",
		minimum:       -128,
		maximum:       127,
		step:          1,
		default_value: 0,
		type:          V4L2_CTRL_TYPE_INTEGER,
		category:      V4L2_CTRL_CAT_VIDEO,
		group:         "Video",
	},
	/* --- audio --- */
	{
		id:            V4L2_CID_AUDIO_MUTE,
		name:          "Mute",
		minimum:       0,
		maximum:       1,
		type:          V4L2_CTRL_TYPE_BOOLEAN,
		category:      V4L2_CTRL_CAT_AUDIO,
		group:         "Audio",
	},{
		id:            V4L2_CID_AUDIO_VOLUME,
		name:          "Volume",
		minimum:       -15,
		maximum:       15,
		step:          1,
		default_value: 0,
		type:          V4L2_CTRL_TYPE_INTEGER,
		category:      V4L2_CTRL_CAT_AUDIO,
		group:         "Audio",
	},
	/* --- private --- */
	{
		id:            V4L2_CID_PRIVATE_INVERT,
		name:          "Invert",
		minimum:       0,
		maximum:       1,
		type:          V4L2_CTRL_TYPE_BOOLEAN,
		group:         "Private",
	},{
		id:            V4L2_CID_PRIVATE_INTERLACE,
		name:          "Interlace",
		minimum:       0,
		maximum:       2,
		type:          V4L2_CTRL_TYPE_MENU,
		group:         "Private",
	},{
		id:            V4L2_CID_PRIVATE_MIRROR,
		name:          "Mirror",
		minimum:       0,
		maximum:       1,
		type:          V4L2_CTRL_TYPE_BOOLEAN,
		group:         "Private",
	},{
		id:            V4L2_CID_PRIVATE_MIRROR,
		name:          "Mirror",
		minimum:       0,
		maximum:       1,
		type:          V4L2_CTRL_TYPE_BOOLEAN,
		group:         "Private",
	},{
		id:            V4L2_CID_PRIVATE_Y_ODD,
		name:          "y offset odd field",
		minimum:       0,
		maximum:       128,
		default_value: 0,
		type:          V4L2_CTRL_TYPE_INTEGER,
		group:         "Private",
	},{
		id:            V4L2_CID_PRIVATE_Y_EVEN,
		name:          "y offset even field",
		minimum:       0,
		maximum:       128,
		default_value: 0,
		type:          V4L2_CTRL_TYPE_INTEGER,
		group:         "Private",
	}
};
const int CTRLS = (sizeof(video_ctrls)/sizeof(struct v4l2_queryctrl));

static const struct v4l2_queryctrl* ctrl_by_id(int id)
{
	int i;
	
	for (i = 0; i < CTRLS; i++)
		if (video_ctrls[i].id == id)
			return video_ctrls+i;
	return NULL;
}

static struct saa7134_format* format_by_fourcc(int  fourcc)
{
	int i;

	for (i = 0; i < FORMATS; i++)
		if (formats[i].fourcc == fourcc)
			return formats+i;
	return NULL;
}

/* ----------------------------------------------------------------------- */
/* resource management                                                     */

static int res_get(struct saa7134_dev *dev, struct saa7134_video_fh *fh, int bit)
{
	if (fh->resources & bit)
		/* have it already allocated */
		return 1;

	/* is it free? */
	down(&dev->lock);
	if (dev->resources & bit) {
		/* no, someone else uses it */
		up(&dev->lock);
		return 0;
	}
	/* it's free, grab it */
	fh->resources  |= bit;
	dev->resources |= bit;
	dprintk("res: get %d\n",bit);
	up(&dev->lock);
	return 1;
}

static
int res_check(struct saa7134_video_fh *fh, int bit)
{
	return (fh->resources & bit);
}

static
int res_locked(struct saa7134_dev *dev, int bit)
{
	return (dev->resources & bit);
}

static
void res_free(struct saa7134_dev *dev, struct saa7134_video_fh *fh, int bits)
{
	if ((fh->resources & bits) != bits)
		BUG();

	down(&dev->lock);
	fh->resources  &= ~bits;
	dev->resources &= ~bits;
	dprintk("res: put %d\n",bits);
	up(&dev->lock);
}

/* ------------------------------------------------------------------ */

static int interlace(struct saa7134_dev *dev, int height)
{
	switch (dev->ctl_interlace) {
	case INTERLACE_AUTO:
		if (height > dev->tvnorm->height/2)
			return TRUE;
		return FALSE;
	case INTERLACE_ON:
		return TRUE;
	case INTERLACE_OFF:
		return FALSE;
	}
	return -1;
}

static void set_tvnorm(struct saa7134_dev *dev, struct saa7134_tvnorm *norm)
{
	struct video_channel c;
	int luma_control,mux;

	dprintk("set tv norm = %s\n",norm->name);
	dev->tvnorm = norm;

	mux = card_in(dev,dev->ctl_input).vmux;
	luma_control = norm->luma_control;
	if (mux > 5)
		luma_control |= 0x80; /* svideo */

	/* setup video decoder */
	saa_writeb(SAA7134_INCR_DELAY,            0x08);
	saa_writeb(SAA7134_ANALOG_IN_CTRL1,       0xc0 | mux);
	saa_writeb(SAA7134_ANALOG_IN_CTRL2,       0x00);

	saa_writeb(SAA7134_ANALOG_IN_CTRL3,       0x90);
	saa_writeb(SAA7134_ANALOG_IN_CTRL4,       0x90);
	saa_writeb(SAA7134_HSYNC_START,           0xeb);
	saa_writeb(SAA7134_HSYNC_STOP,            0xe0);

	saa_writeb(SAA7134_SYNC_CTRL,             norm->sync_control);
	saa_writeb(SAA7134_LUMA_CTRL,             luma_control);
	saa_writeb(SAA7134_DEC_LUMA_BRIGHT,       dev->ctl_bright);
	saa_writeb(SAA7134_DEC_LUMA_CONTRAST,     dev->ctl_contrast);
 
	saa_writeb(SAA7134_DEC_CHROMA_SATURATION, dev->ctl_saturation);
	saa_writeb(SAA7134_DEC_CHROMA_HUE,        dev->ctl_hue);
	saa_writeb(SAA7134_CHROMA_CTRL1,          norm->chroma_ctrl1);
	saa_writeb(SAA7134_CHROMA_GAIN,           norm->chroma_gain);

	saa_writeb(SAA7134_CHROMA_CTRL2,          norm->chroma_ctrl2);
	saa_writeb(SAA7134_MODE_DELAY_CTRL,       0x00);

	saa_writeb(SAA7134_ANALOG_ADC,            0x01);
	saa_writeb(SAA7134_VGATE_START,           0x11);
	saa_writeb(SAA7134_VGATE_STOP,            0xfe);
	saa_writeb(SAA7134_MISC_VGATE_MSB,        0x18); /* FIXME */
	saa_writeb(SAA7134_RAW_DATA_GAIN,         0x40);
	saa_writeb(SAA7134_RAW_DATA_OFFSET,       0x80);

	/* pass down info to the i2c chips (v4l1) */
	memset(&c,0,sizeof(c));
	c.channel = dev->ctl_input;
	switch (norm->id) {
	case V4L2_STD_PAL:   c.norm = VIDEO_MODE_PAL;   break;
	case V4L2_STD_NTSC:  c.norm = VIDEO_MODE_NTSC;  break;
	case V4L2_STD_SECAM: c.norm = VIDEO_MODE_SECAM; break;
	}
	saa7134_i2c_call_clients(dev,VIDIOCSCHAN,&c);
}

static void video_mux(struct saa7134_dev *dev, int input)
{
	dprintk("video input = %d [%s]\n",input,card_in(dev,input).name);
	dev->ctl_input = input;
	set_tvnorm(dev,dev->tvnorm);
	saa7134_tvaudio_setinput(dev,&card_in(dev,input));
}

static void set_h_prescale(struct saa7134_dev *dev, int task, int prescale)
{
	static const struct {
		int xpsc;
		int xacl;
		int xc2_1;
		int xdcg;
		int vpfy;
	} vals[] = {
		/* XPSC XACL XC2_1 XDCG VPFY */
		{    1,   0,    0,    0,   0 },
		{    2,   2,    1,    2,   2 },
		{    3,   4,    1,    3,   2 },
		{    4,   8,    1,    4,   2 },
		{    5,   8,    1,    4,   2 },
		{    6,   8,    1,    4,   3 },
		{    7,   8,    1,    4,   3 },
		{    8,  15,    0,    4,   3 },
		{    9,  15,    0,    4,   3 },
		{   10,  16,    1,    5,   3 },
	};
	static const int count = sizeof(vals)/sizeof(vals[0]);
	int i;

	for (i = 0; i < count; i++)
		if (vals[i].xpsc == prescale)
			break;
	if (i == count)
		return;

	saa_writeb(SAA7134_H_PRESCALE(task), vals[i].xpsc);
	saa_writeb(SAA7134_ACC_LENGTH(task), vals[i].xacl);
	saa_writeb(SAA7134_LEVEL_CTRL(task),
		   (vals[i].xc2_1 << 3) | (vals[i].xdcg));
	saa_andorb(SAA7134_FIR_PREFILTER_CTRL(task), 0x0f,
		   (vals[i].vpfy << 2) | vals[i].vpfy);
}

static void set_v_scale(struct saa7134_dev *dev, int task, int yscale)
{
	int val,mirror;
	
	saa_writeb(SAA7134_V_SCALE_RATIO1(task), yscale &  0xff);
	saa_writeb(SAA7134_V_SCALE_RATIO2(task), yscale >> 8);

	mirror = (dev->ctl_mirror) ? 0x02 : 0x00;
	if (yscale < 2048) {
		/* LPI */
		dprintk("yscale LPI yscale=%d\n",yscale);
		saa_writeb(SAA7134_V_FILTER(task), 0x00 | mirror);
		saa_writeb(SAA7134_LUMA_CONTRAST(task), 0x40);
		saa_writeb(SAA7134_CHROMA_SATURATION(task), 0x40);
	} else {
		/* ACM */
		val = 0x40 * 1024 / yscale;
		dprintk("yscale ACM yscale=%d val=0x%x\n",yscale,val);
		saa_writeb(SAA7134_V_FILTER(task), 0x01 | mirror);
		saa_writeb(SAA7134_LUMA_CONTRAST(task), val);
		saa_writeb(SAA7134_CHROMA_SATURATION(task), val);
	}
	saa_writeb(SAA7134_LUMA_BRIGHT(task),       0x80);
}

static void set_size(struct saa7134_dev *dev, int task,
		     int width, int height, int interlace)
{
	struct saa7134_tvnorm *norm = dev->tvnorm;
	int prescale,xscale,yscale,y_even,y_odd;
	int div = interlace ? 2 : 1;

	/* setup video scaler */
	saa_writeb(SAA7134_VIDEO_H_START1(task), norm->h_start       &  0xff);
	saa_writeb(SAA7134_VIDEO_H_START2(task), norm->h_start       >> 8);
	saa_writeb(SAA7134_VIDEO_H_STOP1(task),  norm->h_stop        &  0xff);
	saa_writeb(SAA7134_VIDEO_H_STOP2(task),  norm->h_stop        >> 8);
	saa_writeb(SAA7134_VIDEO_V_START1(task), norm->video_v_start &  0xff);
	saa_writeb(SAA7134_VIDEO_V_START2(task), norm->video_v_start >> 8);
	saa_writeb(SAA7134_VIDEO_V_STOP1(task),  norm->video_v_stop  &  0xff);
	saa_writeb(SAA7134_VIDEO_V_STOP2(task),  norm->video_v_stop  >> 8);

	prescale = norm->width / width;
	if (0 == prescale)
		prescale = 1;
	xscale = 1024 * norm->width / prescale / width;
	yscale = 512 * div * norm->height / height;
       	dprintk("prescale=%d xscale=%d yscale=%d\n",prescale,xscale,yscale);
	set_h_prescale(dev,task,prescale);
	saa_writeb(SAA7134_H_SCALE_INC1(task),      xscale &  0xff);
	saa_writeb(SAA7134_H_SCALE_INC2(task),      xscale >> 8);
	set_v_scale(dev,task,yscale);
	
	saa_writeb(SAA7134_VIDEO_PIXELS1(task),     width  & 0xff);
	saa_writeb(SAA7134_VIDEO_PIXELS2(task),     width  >> 8);
	saa_writeb(SAA7134_VIDEO_LINES1(task),      height/div & 0xff);
	saa_writeb(SAA7134_VIDEO_LINES2(task),      height/div >> 8);

	/* deinterlace y offsets */
	if (interlace) {
		y_odd  = dev->ctl_y_odd;
		y_even = dev->ctl_y_even + yscale / 32;
		saa_writeb(SAA7134_V_PHASE_OFFSET0(task), y_odd);
		saa_writeb(SAA7134_V_PHASE_OFFSET1(task), y_even);
		saa_writeb(SAA7134_V_PHASE_OFFSET2(task), y_odd);
		saa_writeb(SAA7134_V_PHASE_OFFSET3(task), y_even);
	} else {
		y_odd  = dev->ctl_y_odd;
		y_even = dev->ctl_y_even + yscale / 64;
		saa_writeb(SAA7134_V_PHASE_OFFSET0(task), y_odd);
		saa_writeb(SAA7134_V_PHASE_OFFSET1(task), y_even);
		saa_writeb(SAA7134_V_PHASE_OFFSET2(task), y_odd);
		saa_writeb(SAA7134_V_PHASE_OFFSET3(task), y_even);
	}
}

/* ------------------------------------------------------------------ */

struct cliplist {
	__u16 position;
	__u8  enable;
	__u8  disable;
};

static void sort_cliplist(struct cliplist *cl, int entries)
{
	struct cliplist swap;
	int i,j,n;
	
	for (i = entries-2; i >= 0; i--) {
		for (n = 0, j = 0; j <= i; j++) {
			if (cl[j].position > cl[j+1].position) {
				swap = cl[j];
				cl[j] = cl[j+1];
				cl[j+1] = swap;
				n++;
			}
		}
		if (0 == n)
			break;
	}
}

static void set_cliplist(struct saa7134_dev *dev, int reg,
			struct cliplist *cl, int entries, char *name)
{
	__u8 winbits = 0;
	int i;

	for (i = 0; i < entries; i++) {
		winbits |= cl[i].enable;
		winbits &= ~cl[i].disable;
		if (i < 15 && cl[i].position == cl[i+1].position)
			continue;
		saa_writeb(reg + 0, winbits);
		saa_writeb(reg + 2, cl[i].position & 0xff);
		saa_writeb(reg + 3, cl[i].position >> 8);
		dprintk("clip: %s winbits=%02x pos=%d\n",
			name,winbits,cl[i].position);
		reg += 8;
	}
	for (; reg < 0x400; reg += 8) {
		saa_writeb(reg+ 0, 0);
		saa_writeb(reg + 1, 0);
		saa_writeb(reg + 2, 0);
		saa_writeb(reg + 3, 0);
	}
}

static int clip_range(int val)
{
	if (val < 0)
		val = 0;
	return val;
}

static int setup_clipping(struct saa7134_dev *dev, struct v4l2_clip *clips,
			  int nclips, int interlace)
{
	struct cliplist col[16], row[16];
	int cols, rows, i;
	int div = interlace ? 2 : 1;

	memset(col,0,sizeof(col)); cols = 0;
	memset(row,0,sizeof(row)); rows = 0;
	for (i = 0; i < nclips && i < 8; i++) {
		col[cols].position = clip_range(clips[i].x);
		col[cols].enable   = (1 << i);
		cols++;
		col[cols].position = clip_range(clips[i].x+clips[i].width);
		col[cols].disable  = (1 << i);
		cols++;
		row[rows].position = clip_range(clips[i].y / div);
		row[rows].enable   = (1 << i);
		rows++;
		row[rows].position = clip_range((clips[i].y + clips[i].height)
						/ div);
		row[rows].disable  = (1 << i);
		rows++;
	}
	sort_cliplist(col,cols);
	sort_cliplist(row,rows);
	set_cliplist(dev,0x380,col,cols,"cols");
	set_cliplist(dev,0x384,row,rows,"rows");
	return 0;
}

static int start_preview(struct saa7134_dev *dev, struct saa7134_video_fh *fh)
{
	unsigned long base,control,bpl;
	
	if (fh->win.width  <  48  ||
	    fh->win.height <  32  ||
	    fh->win.width  > dev->tvnorm->width  ||
	    fh->win.height > dev->tvnorm->height ||
	    NULL == dev->ovfmt    ||
	    NULL == dev->ovbuf.base[0])
		return -EINVAL;
	
	dev->ovinterlace = interlace(dev,fh->win.height);
	dprintk("start_preview %dx%d+%d+%d %s interlace=%d\n",
		fh->win.width,fh->win.height,fh->win.x,fh->win.y,
		dev->ovfmt->name,dev->ovinterlace);

	/* setup window + clipping */
	set_size(dev,TASK_B,fh->win.width,fh->win.height,dev->ovinterlace);
	setup_clipping(dev,fh->clips,fh->nclips,dev->ovinterlace);
	if (dev->ovfmt->yuv)
		saa_andorb(SAA7134_DATA_PATH(TASK_B), 0x3f, 0x03);
	else
		saa_andorb(SAA7134_DATA_PATH(TASK_B), 0x3f, 0x01);
	saa_writeb(SAA7134_OFMT_VIDEO_B, dev->ovfmt->pm | 0x20);

	/* dma: setup channel 1 (= Video Task B) */
	base  = (unsigned long)dev->ovbuf.base[0];
	base += dev->ovbuf.fmt.bytesperline * fh->win.y;
	base += dev->ovbuf.fmt.depth/8      * fh->win.x;
	bpl   = dev->ovbuf.fmt.bytesperline;
	control = SAA7134_RS_CONTROL_BURST_16;
	if (dev->ovfmt->bswap)
		control |= SAA7134_RS_CONTROL_BSWAP;
	if (dev->ovfmt->wswap)
		control |= SAA7134_RS_CONTROL_WSWAP;
	if (dev->ovinterlace) {
		saa_writel(SAA7134_RS_BA1(1),base);
		saa_writel(SAA7134_RS_BA2(1),base+bpl);
		saa_writel(SAA7134_RS_PITCH(1),bpl*2);
		saa_writel(SAA7134_RS_CONTROL(1),control);
	} else {
		saa_writel(SAA7134_RS_BA1(1),base);
		saa_writel(SAA7134_RS_BA2(1),base);
		saa_writel(SAA7134_RS_PITCH(1),bpl);
		saa_writel(SAA7134_RS_CONTROL(1),control);
	}

	/* start dma */
	dev->ovenable = 1;
	saa7134_set_dmabits(dev);

	return 0;
}

static int stop_preview(struct saa7134_dev *dev, struct saa7134_video_fh *fh)
{
	dev->ovenable = 0;
	saa7134_set_dmabits(dev);
	return 0;
}

/* ------------------------------------------------------------------ */

static int buffer_activate(struct saa7134_dev *dev, struct saa7134_buf *buf)
{
	unsigned long base,control,bpl;
	unsigned long bpl_uv,lines_uv,base2,base3; /* planar */

	dprintk("buffer_activate buf=%p\n",buf);
	buf->vb.state = STATE_ACTIVE;
	
	set_size(dev,TASK_A,buf->vb.width,buf->vb.height,
		 (buf->vb.field & VBUF_FIELD_INTER));
	if (buf->fmt->yuv)
		saa_andorb(SAA7134_DATA_PATH(TASK_A), 0x3f, 0x03);
	else
		saa_andorb(SAA7134_DATA_PATH(TASK_A), 0x3f, 0x01);
	saa_writeb(SAA7134_OFMT_VIDEO_A, buf->fmt->pm);

	/* DMA: setup channel 0 (= Video Task A0) */
	base = buf->pt.offset;
	if (buf->fmt->planar)
		bpl = buf->vb.width;
	else
		bpl = (buf->vb.width * buf->fmt->depth) / 8;
	control = SAA7134_RS_CONTROL_BURST_16 |
		SAA7134_RS_CONTROL_ME |
		(buf->pt.dma >> 12);
	if (buf->fmt->bswap)
		control |= SAA7134_RS_CONTROL_BSWAP;
	if (buf->fmt->wswap)
		control |= SAA7134_RS_CONTROL_WSWAP;
	if (VBUF_FIELD_INTER & buf->vb.field) {
		/* interlaced */
		saa_writel(SAA7134_RS_BA1(0),base);
		saa_writel(SAA7134_RS_BA2(0),base+bpl);
		saa_writel(SAA7134_RS_PITCH(0),bpl*2);
	} else {
		/* non-interlaced */
		saa_writel(SAA7134_RS_BA1(0),base);
		saa_writel(SAA7134_RS_BA2(0),base);
		saa_writel(SAA7134_RS_PITCH(0),bpl);
	}
	saa_writel(SAA7134_RS_CONTROL(0),control);

	if (buf->fmt->planar) {
		/* DMA: setup channel 4+5 (= planar task A) */
		bpl_uv   = bpl >> buf->fmt->hshift;
		lines_uv = buf->vb.height >> buf->fmt->vshift;
		base2    = base + bpl * buf->vb.height;
		base3    = base2 + bpl_uv * lines_uv;
		dprintk("uv: bpl=%ld lines=%ld base2/3=%ld/%ld\n",
			bpl_uv,lines_uv,base2,base3);
		if (VBUF_FIELD_INTER & buf->vb.field) {
			/* interlaced */
			saa_writel(SAA7134_RS_BA1(4),base2);
			saa_writel(SAA7134_RS_BA2(4),base2+bpl_uv);
			saa_writel(SAA7134_RS_PITCH(4),bpl_uv*2);
			saa_writel(SAA7134_RS_BA1(5),base3);
			saa_writel(SAA7134_RS_BA2(5),base3+bpl_uv);
			saa_writel(SAA7134_RS_PITCH(5),bpl_uv*2);
		} else {
			/* non-interlaced */
			saa_writel(SAA7134_RS_BA1(4),base2);
			saa_writel(SAA7134_RS_BA2(4),base2);
			saa_writel(SAA7134_RS_PITCH(4),bpl_uv);
			saa_writel(SAA7134_RS_BA1(5),base3);
			saa_writel(SAA7134_RS_BA2(5),base3);
			saa_writel(SAA7134_RS_PITCH(5),bpl_uv);
		}
		saa_writel(SAA7134_RS_CONTROL(4),control);
		saa_writel(SAA7134_RS_CONTROL(5),control);
	}

	/* start DMA */
	saa7134_set_dmabits(dev);
	mod_timer(&dev->video_q.timeout, jiffies+BUFFER_TIMEOUT);
	return 0;
}

static int buffer_prepare(struct file *file, struct videobuf_buffer *vb, int field)
{
	static const char *fname[4] = { "Oops", "even", "odd", "both" };
	struct saa7134_video_fh *fh = file->private_data;
	struct saa7134_dev *dev = fh->dev;
	struct saa7134_buf *buf = (struct saa7134_buf *)vb;
	int size,err;
	
	/* sanity checks */
	if (NULL == fh->fmt)
		return -EINVAL;
	if (fh->width  < 48 ||
	    fh->height < 32 ||
	    fh->width  > dev->tvnorm->width ||
	    fh->height > dev->tvnorm->height)
		return -EINVAL;
	size = (fh->width * fh->height * fh->fmt->depth) >> 3;
	if (0 != buf->vb.baddr  &&  buf->vb.bsize < size)
		return -EINVAL;
	if (0 == field)
		field = fh->inter ? VBUF_FIELD_INTER : VBUF_FIELD_ODD;

	dprintk("buffer_prepare [size=%dx%d,bytes=%d,fields=%s,%s]\n",
		fh->width,fh->height,size,fname[field],fh->fmt->name);
	if (buf->vb.width  != fh->width  ||
	    buf->vb.height != fh->height ||
	    buf->vb.size   != size       ||
	    buf->fmt       != fh->fmt) {
		saa7134_dma_free(dev,buf);
	}

	if (STATE_NEEDS_INIT == buf->vb.state) {
		buf->vb.width  = fh->width;
		buf->vb.height = fh->height;
		buf->vb.size   = size;
		buf->fmt       = fh->fmt;

		err = videobuf_iolock(dev->pci,&buf->vb);
		if (err)
			goto oops;
		err = saa7134_pgtable_build(dev->pci,&buf->pt,
					    buf->vb.dma.sglist,
					    buf->vb.dma.sglen);
		if (err)
			goto oops;
	}
	buf->vb.state = STATE_PREPARED;
	buf->odd_seen = 0;
	buf->activate = buffer_activate;
	buf->vb.field = field;
	return 0;

 oops:
	saa7134_dma_free(dev,buf);
	return err;
}

static int
buffer_setup(struct file *file, int *count, int *size)
{
	struct saa7134_video_fh *fh = file->private_data;

	*size = fh->fmt->depth * fh->width * fh->height >> 3;
	if (0 == *count)
		*count = gbuffers;
	while (*size * *count > gbuffers * gbufsize)
		(*count)--;
	return 0;
}

static void buffer_queue(struct file *file, struct videobuf_buffer *vb)
{
	struct saa7134_video_fh *fh = file->private_data;
	struct saa7134_buf *buf = (struct saa7134_buf *)vb;
	
	saa7134_buffer_queue(fh->dev,&fh->dev->video_q,buf);
}

static void buffer_release(struct file *file, struct videobuf_buffer *vb)
{
	struct saa7134_video_fh *fh = file->private_data;
	struct saa7134_buf *buf = (struct saa7134_buf *)vb;
	
	saa7134_dma_free(fh->dev,buf);
}

static struct videobuf_queue_ops video_qops = {
	buf_setup:    buffer_setup,
	buf_prepare:  buffer_prepare,
	buf_queue:    buffer_queue,
	buf_release:  buffer_release,
};

/* ------------------------------------------------------------------ */

static int get_control(struct saa7134_dev *dev, struct v4l2_control *c)
{
	const struct v4l2_queryctrl* ctrl;

	ctrl = ctrl_by_id(c->id);
	if (NULL == ctrl)
		return -EINVAL;
	switch (c->id) {
	case V4L2_CID_BRIGHTNESS:
		c->value = dev->ctl_bright;
		break;
	case V4L2_CID_HUE:
		c->value = dev->ctl_hue;
		break;
	case V4L2_CID_CONTRAST:
		c->value = dev->ctl_contrast;
		break;
	case V4L2_CID_SATURATION:
		c->value = dev->ctl_saturation;
		break;
	case V4L2_CID_AUDIO_MUTE:
		c->value = dev->ctl_mute;
		break;
	case V4L2_CID_AUDIO_VOLUME:
		c->value = dev->ctl_volume;
		break;
	case V4L2_CID_PRIVATE_INVERT:
		c->value = dev->ctl_invert;
		break;
	case V4L2_CID_PRIVATE_INTERLACE:
		c->value = dev->ctl_interlace;
		break;
	case V4L2_CID_PRIVATE_MIRROR:
		c->value = dev->ctl_mirror;
		break;
	case V4L2_CID_PRIVATE_Y_EVEN:
		c->value = dev->ctl_y_even;
		break;
	case V4L2_CID_PRIVATE_Y_ODD:
		c->value = dev->ctl_y_odd;
		break;
	default:
		return -EINVAL;
	}
	return 0;
}

static int set_control(struct saa7134_dev *dev, struct saa7134_video_fh *fh,
		       struct v4l2_control *c)
{
	const struct v4l2_queryctrl* ctrl;
	unsigned long flags;
	int restart_overlay = 0;

	ctrl = ctrl_by_id(c->id);
	if (NULL == ctrl)
		return -EINVAL;
	dprintk("set_control name=%s val=%d\n",ctrl->name,c->value);
	switch (ctrl->type) {
	case V4L2_CTRL_TYPE_BOOLEAN:
	case V4L2_CTRL_TYPE_MENU:
	case V4L2_CTRL_TYPE_INTEGER:
		if (c->value < ctrl->minimum)
			c->value = ctrl->minimum;
		if (c->value > ctrl->maximum)
			c->value = ctrl->maximum;
		break;
	};
	switch (c->id) {
	case V4L2_CID_BRIGHTNESS:
		dev->ctl_bright = c->value;
		saa_writeb(SAA7134_DEC_LUMA_BRIGHT, dev->ctl_bright);
		break;
	case V4L2_CID_HUE:
		dev->ctl_hue = c->value;
		saa_writeb(SAA7134_DEC_CHROMA_HUE, dev->ctl_hue);
		break;
	case V4L2_CID_CONTRAST:
		dev->ctl_contrast = c->value;
		saa_writeb(SAA7134_DEC_LUMA_CONTRAST,
			   dev->ctl_invert ? -dev->ctl_contrast : dev->ctl_contrast);
		break;
	case V4L2_CID_SATURATION:
		dev->ctl_saturation = c->value;
		saa_writeb(SAA7134_DEC_CHROMA_SATURATION,
			   dev->ctl_invert ? -dev->ctl_saturation : dev->ctl_saturation);
		break;
	case V4L2_CID_AUDIO_MUTE:
		dev->ctl_mute = c->value;
		saa7134_tvaudio_setmute(dev);
		break;
	case V4L2_CID_AUDIO_VOLUME:
		dev->ctl_volume = c->value;
		saa7134_tvaudio_setvolume(dev,dev->ctl_volume);
		break;
	case V4L2_CID_PRIVATE_INVERT:
		dev->ctl_invert = c->value;
		saa_writeb(SAA7134_DEC_LUMA_CONTRAST,
			   dev->ctl_invert ? -dev->ctl_contrast : dev->ctl_contrast);
		saa_writeb(SAA7134_DEC_CHROMA_SATURATION,
			   dev->ctl_invert ? -dev->ctl_saturation : dev->ctl_saturation);
		break;
	case V4L2_CID_PRIVATE_INTERLACE:
		dev->ctl_interlace = c->value;
		restart_overlay = 1;
		break;
	case V4L2_CID_PRIVATE_MIRROR:
		dev->ctl_mirror = c->value;
		restart_overlay = 1;
		break;
	case V4L2_CID_PRIVATE_Y_EVEN:
		dev->ctl_y_even = c->value;
		restart_overlay = 1;
		break;
	case V4L2_CID_PRIVATE_Y_ODD:
		dev->ctl_y_odd = c->value;
		restart_overlay = 1;
		break;
	default:
		return -EINVAL;
	}
	if (restart_overlay && res_check(fh, RESOURCE_OVERLAY)) {
		spin_lock_irqsave(&dev->slock,flags);
		stop_preview(dev,fh);
		start_preview(dev,fh);
		spin_unlock_irqrestore(&dev->slock,flags);
	}
	return 0;
}

/* ------------------------------------------------------------------ */

int saa7134_common_ioctl(struct saa7134_dev *dev,
			 unsigned int cmd, void *arg)
{
	int err;

	switch (cmd) {
		
	/* --- tv standards ------------------------------------------ */
	case VIDIOC_ENUMSTD:
	{
		struct v4l2_enumstd *e = arg;

		if (e->index < 0 || e->index >= TVNORMS)
			return -EINVAL;
		err = v4l2_video_std_construct(&e->std, tvnorms[e->index].id,
					       0);
		if (err < 0)
			return err;
		e->inputs  = 0x01f;
		return 0;
	}
	case VIDIOC_G_STD:
	{
		struct v4l2_standard *s = arg;
		v4l2_video_std_construct(s,dev->tvnorm->id,0);
		return 0;
	}
	case VIDIOC_S_STD:
	{
		struct v4l2_standard *s = arg;
		int i, id = v4l2_video_std_confirm(s);

		for(i = 0; i < TVNORMS; i++)
			if (id == tvnorms[i].id)
				break;
		if (i == TVNORMS)
			return -EINVAL;

		down(&dev->lock);
#if 0 /* FIXME */
		if (res_check(fh, RESOURCE_OVERLAY)) {
			spin_lock_irqsave(&dev->slock,flags);
			stop_preview(dev,fh);
			set_tvnorm(dev,&tvnorms[i]);
			start_preview(dev,fh);
			spin_unlock_irqrestore(&dev->slock,flags);
		} else 
#endif
			set_tvnorm(dev,&tvnorms[i]);
		
		up(&dev->lock);
		return 0;
	}

	/* --- input switching --------------------------------------- */
	case VIDIOC_ENUMINPUT:
	{
		struct v4l2_input *i = arg;
		
		if (i->index >= SAA7134_INPUT_MAX)
			return -EINVAL;
		if (NULL == card_in(dev,i->index).name)
			return -EINVAL;
		i->type = V4L2_INPUT_TYPE_CAMERA;
		i->capability = 0;
		i->assoc_audio = 0;
		strcpy(i->name,card_in(dev,i->index).name);
		if (card_in(dev,i->index).tv) {
			i->type = V4L2_INPUT_TYPE_TUNER;
			i->capability = V4L2_INPUT_CAP_AUDIO;
		}
		return 0;
	}
	case VIDIOC_G_INPUT:
	{
		int *i = arg;
		*i = dev->ctl_input;
		return 0;
	}
	case VIDIOC_S_INPUT:
	{
		int *i = arg;
		
		if (*i < 0  ||  *i >= SAA7134_INPUT_MAX)
			return -EINVAL;
		if (NULL == card_in(dev,*i).name)
			return -EINVAL;
		down(&dev->lock);
		video_mux(dev,*i);
		up(&dev->lock);
		return 0;
	}

	/* --- tuner ioctls ------------------------------------------ */
	case VIDIOC_G_TUNER:
	{
		struct v4l2_tuner *t = arg;
		int n;

		memset(t,0,sizeof(*t));
		for (n = 0; n < SAA7134_INPUT_MAX; n++)
			if (card_in(dev,n).tv)
				break;
		if (NULL != card_in(dev,n).name) {
			t->input = n;
			strcpy(t->name, "Television");
			v4l2_video_std_construct(&t->std,dev->tvnorm->id,0);
			t->capability = V4L2_TUNER_CAP_NORM |
				V4L2_TUNER_CAP_STEREO |
				V4L2_TUNER_CAP_LANG1 |
				V4L2_TUNER_CAP_LANG2;
			t->rangehigh = 0xffffffffUL;
			if (dev->tvaudio) {
				t->rxsubchans = saa7134_tvaudio_getstereo
					(dev,dev->tvaudio);
			} else {
				t->rxsubchans = V4L2_TUNER_SUB_MONO;
			}
#if 1
			/* fill audmode -- FIXME: allow manual switching */
			t->audmode = V4L2_TUNER_MODE_MONO;
			if (t->rxsubchans & V4L2_TUNER_SUB_STEREO)
				t->audmode = V4L2_TUNER_MODE_STEREO;
			else if (t->rxsubchans & V4L2_TUNER_SUB_LANG1)
				t->audmode = V4L2_TUNER_MODE_LANG1;
			else if (t->rxsubchans & V4L2_TUNER_SUB_LANG2)
				t->audmode = V4L2_TUNER_MODE_LANG2;
#endif
		}
		if (0 != (saa_readb(SAA7134_STATUS_VIDEO1) & 0x03))
			t->signal = 0xffff;
		return 0;
	}
	case VIDIOC_S_TUNER:
	{
#if 0
		struct v4l2_tuner *t = arg;
#endif
		return 0;
	}
	case VIDIOC_G_FREQ:
	{
		unsigned long *freq = arg;
		*freq = dev->ctl_freq;
		return 0;
	}
	case VIDIOC_S_FREQ:
	{
		unsigned long *freq = arg;

		down(&dev->lock);
		dev->ctl_freq=*freq;
		saa7134_i2c_call_clients(dev,VIDIOCSFREQ,freq);
		up(&dev->lock);
		return 0;
	}
	default:
		return -EINVAL;
	}
}

/* ------------------------------------------------------------------ */

static int video_open(struct inode *inode, struct file *file)
{
	unsigned int minor = minor(inode->i_rdev);
	struct saa7134_dev *h,*dev = NULL;
	struct saa7134_video_fh *fh;
	struct list_head *list;
	int radio = 0;
	
	list_for_each(list,&saa7134_devlist) {
		h = list_entry(list, struct saa7134_dev, devlist);
		if (h->video_dev.minor == minor)
			dev = h;
		if (h->radio_dev.minor == minor) {
			radio = 1;
			dev = h;
		}
	}
	if (NULL == dev)
		return -ENODEV;

	dprintk("open minor=%d radio=%d\n",minor,radio);

	/* allocate + initialize per filehandle data */
	fh = kmalloc(sizeof(*fh),GFP_KERNEL);
	if (NULL == fh)
		return -ENOMEM;
	memset(fh,0,sizeof(*fh));
	file->private_data = fh;
	fh->dev      = dev;
	fh->radio    = radio;
	fh->fmt      = format_by_fourcc(V4L2_PIX_FMT_BGR24);
	fh->width    = 320;
	fh->height   = 240;
	videobuf_queue_init(&fh->q, &video_qops, dev->pci, &dev->slock,
			    V4L2_BUF_TYPE_CAPTURE,
			    sizeof(struct saa7134_buf));
	
        init_MUTEX(&fh->q.lock);
	if (fh->radio) {
		/* switch to radio mode */
		saa7134_tvaudio_setinput(dev,&card(dev).radio);
		saa7134_i2c_call_clients(dev,AUDC_SET_RADIO,NULL);
	} else {
		/* switch to TV mode */
		set_tvnorm(dev,dev->tvnorm);
		video_mux(dev,dev->ctl_input);
	}
        return 0;
}

static ssize_t
video_read(struct file *file, char *data, size_t count, loff_t *ppos)
{
	struct saa7134_video_fh *fh = file->private_data;

	return videobuf_read_one(file, &fh->q, data, count, ppos);
}

static unsigned int
video_poll(struct file *file, struct poll_table_struct *wait)
{
	struct saa7134_video_fh *fh = file->private_data;
	struct videobuf_buffer *buf = NULL;

	if (res_check(fh,RESOURCE_STREAMING)) {
		if (!list_empty(&fh->q.stream))
			buf = list_entry(fh->q.stream.next, struct videobuf_buffer, stream);
	} else {
		down(&fh->q.lock);
		if (-1 == fh->q.read_off) {
                        /* need to capture a new frame */
			if (res_locked(fh->dev,RESOURCE_STREAMING)) {
                                up(&fh->q.lock);
                                return POLLERR;
                        }
                        if (0 != fh->q.ops->buf_prepare(file,fh->q.read_buf,0)) {
                                up(&fh->q.lock);
                                return POLLERR;
                        }
                        fh->q.ops->buf_queue(file,fh->q.read_buf);
                        fh->q.read_off = 0;
		}
		up(&fh->q.lock);
		buf = fh->q.read_buf;
	}

	if (!buf)
		return POLLERR;

	poll_wait(file, &buf->done, wait);
	if (buf->state == STATE_DONE ||
	    buf->state == STATE_ERROR)
		return POLLIN|POLLRDNORM;
	return 0;
}

static int video_release(struct inode *inode, struct file *file)
{
	struct saa7134_video_fh  *fh  = file->private_data;
	struct saa7134_dev *dev = fh->dev;
	unsigned long flags;

	if (res_check(fh, RESOURCE_OVERLAY)) {
		spin_lock_irqsave(&dev->slock,flags);
		stop_preview(dev,fh);
		spin_unlock_irqrestore(&dev->slock,flags);
		res_free(dev,fh,RESOURCE_OVERLAY);
	}
	if (res_check(fh, RESOURCE_STREAMING)) {
		videobuf_queue_cancel(file,&fh->q);
		res_free(dev,fh,RESOURCE_STREAMING);
	}
	if (fh->q.read_buf) {
		buffer_release(file,fh->q.read_buf);
		kfree(fh->q.read_buf);
	}
	
	file->private_data = NULL;
	kfree(fh);
	return 0;
}

static int
video_mmap(struct file *file, struct vm_area_struct * vma)
{
	struct saa7134_video_fh *fh = file->private_data;
	
	return videobuf_mmap_mapper(vma,&fh->q);
}

/*
 * This function is _not_ called directly, but from
 * video_generic_ioctl (and maybe others).  userspace
 * copying is done already, arg is a kernel pointer.
 */
static int video_do_ioctl(struct inode *inode, struct file *file,
			  unsigned int cmd, void *arg)
{
	struct saa7134_video_fh *fh = file->private_data;
	struct saa7134_dev *dev = fh->dev;
	unsigned long flags;
	int err;

	if (video_debug > 1)
		saa7134_print_ioctl(dev->name,cmd);
	switch (cmd) {
	case VIDIOC_QUERYCAP:
	{
		struct v4l2_capability *cap = arg;
		int n;

		memset(cap,0,sizeof(*cap));
                strcpy(cap->name, dev->name);
		cap->type = V4L2_TYPE_CAPTURE;
		cap->flags = V4L2_FLAG_PREVIEW | V4L2_FLAG_TUNER |
			V4L2_FLAG_READ | V4L2_FLAG_STREAMING |
			V4L2_FLAG_SELECT;
		for (n = 0; n < SAA7134_INPUT_MAX; n++)
			if (NULL == card_in(dev,n).name)
				break;
		cap->inputs = n;
		cap->outputs = 0;
		cap->audios = 1;
		cap->maxwidth = dev->tvnorm->width;
		cap->maxheight = dev->tvnorm->height;
		cap->minwidth = 48;
		cap->minheight = 32;
		cap->maxframerate = 30;
		return 0;
	}

	/* --- common ioctls ---------------------------------------- */
	case VIDIOC_ENUMSTD:
	case VIDIOC_G_STD:
	case VIDIOC_S_STD:
	case VIDIOC_ENUMINPUT:
	case VIDIOC_G_INPUT:
	case VIDIOC_S_INPUT:
	case VIDIOC_G_TUNER:
	case VIDIOC_S_TUNER:
	case VIDIOC_G_FREQ:
	case VIDIOC_S_FREQ:
		return saa7134_common_ioctl(dev,cmd,arg);
		
	/* --- control ioctls ---------------------------------------- */
	case VIDIOC_QUERYCTRL:
	{
		const struct v4l2_queryctrl *ctrl;
		struct v4l2_queryctrl *c = arg;

		v4l2_fill_ctrl_category(c);
		if ((c->id <  V4L2_CID_BASE ||
		     c->id >= V4L2_CID_LASTP1) &&
		    (c->id <  V4L2_CID_PRIVATE_BASE ||
		     c->id >= V4L2_CID_PRIVATE_LASTP1))
			return -EINVAL;
		ctrl = ctrl_by_id(c->id);
		*c = (NULL != ctrl) ? *ctrl : no_ctrl;
		return 0;
	}
	case VIDIOC_QUERYMENU:
	{
		struct v4l2_querymenu *qm = arg;
		
		if (V4L2_CID_PRIVATE_INTERLACE == qm->id) {
			switch (qm->index) {
			case INTERLACE_AUTO:
				strcpy(qm->name, "auto");
				break;
			case INTERLACE_ON:
				strcpy(qm->name, "on");
				break;
			case INTERLACE_OFF:
				strcpy(qm->name, "off");
				break;
			default: return -EINVAL;
			}

		} else {
			return -EINVAL;
		}
		return 0;
	}
	case VIDIOC_G_CTRL:
		return get_control(dev,arg);
	case VIDIOC_S_CTRL:
	{
		down(&dev->lock);
		err = set_control(dev,fh,arg);
		up(&dev->lock);
		return err;
	}
	case VIDIOC_G_AUDIO:
	{
		struct v4l2_audio *a = arg;

		memset(a,0,sizeof(*a));
		strcpy(a->name,"audio");
		return 0;
	}
	case VIDIOC_S_AUDIO:
		return 0;
        case VIDIOC_G_PARM:
        {
                struct v4l2_captureparm *parm = arg;
                memset(parm,0,sizeof(*parm));
                return 0;
        }

	/* --- preview ioctls ---------------------------------------- */
	case VIDIOC_ENUM_FBUFFMT:
	case VIDIOC_ENUM_PIXFMT:
	{
		struct v4l2_fmtdesc *f = arg;
		int index;

		index = f->index;
		if (index < 0 || index >= FORMATS)
			return -EINVAL;
		memset(f,0,sizeof(*f));
		f->index = index;
		strncpy(f->description,formats[index].name,31);
		f->pixelformat = formats[index].fourcc;
		f->depth = formats[index].depth;
		return 0;
	}
	case VIDIOC_G_FBUF:
	{
		struct v4l2_framebuffer *fb = arg;

		*fb = dev->ovbuf;
		fb->capability = V4L2_FBUF_CAP_CLIPPING;
		return 0;
	}
	case VIDIOC_S_FBUF:
	{
		struct v4l2_framebuffer *fb = arg;
		struct saa7134_format *fmt;
		
		if(!capable(CAP_SYS_ADMIN) &&
		   !capable(CAP_SYS_RAWIO))
			return -EPERM;

		/* check args */
		fmt = format_by_fourcc(fb->fmt.pixelformat);
		if (NULL == fmt)
			return -EINVAL;

		/* ok, accept it */
		dev->ovbuf = *fb;
		dev->ovfmt = fmt;
		if (!(dev->ovbuf.fmt.flags & V4L2_FMT_FLAG_BYTESPERLINE)) {
			dev->ovbuf.fmt.bytesperline =
				dev->ovbuf.fmt.width*fmt->depth/8;
			dev->ovbuf.fmt.flags |= V4L2_FMT_FLAG_BYTESPERLINE;
		}
		return 0;
	}
	case VIDIOC_G_WIN:
	{
		struct v4l2_window *win = arg;

		*win = fh->win;
		return 0;
	}
	case VIDIOC_S_WIN:
	{
		struct v4l2_window *win = arg;

		if (win->width  <  48 ||
		    win->height <  32 ||
		    win->width  > dev->tvnorm->width  ||
		    win->height > dev->tvnorm->height ||
		    NULL == dev->ovfmt)
			return -EINVAL;
		if (win->clipcount > 2048)
			return -EINVAL;

		down(&dev->lock);
		fh->win = *win;
		fh->nclips = win->clipcount;
		if (fh->nclips > 8)
			fh->nclips = 8;
		if (copy_from_user(fh->clips,win->clips,
				   sizeof(struct v4l2_clip)*fh->nclips)) {
			up(&dev->lock);
			return -EFAULT;
		}

		if (res_check(fh, RESOURCE_OVERLAY)) {
			spin_lock_irqsave(&dev->slock,flags);
			stop_preview(dev,fh);
			start_preview(dev,fh);
			spin_unlock_irqrestore(&dev->slock,flags);
		}
		up(&dev->lock);
		return 0;
	}
	case VIDIOC_PREVIEW:
	{
		int *on = arg;

		if (*on) {
			if (!res_get(dev,fh,RESOURCE_OVERLAY))
				return -EBUSY;
			spin_lock_irqsave(&dev->slock,flags);
			start_preview(dev,fh);
			spin_unlock_irqrestore(&dev->slock,flags);
		}
		if (!*on) {
			if (!res_check(fh, RESOURCE_OVERLAY))
				return -EINVAL;
			spin_lock_irqsave(&dev->slock,flags);
			stop_preview(dev,fh);
			spin_unlock_irqrestore(&dev->slock,flags);
			res_free(dev,fh,RESOURCE_OVERLAY);
		}
		return 0;
	}

	/* --- capture ioctls ---------------------------------------- */
	case VIDIOC_G_FMT:
	{
		struct v4l2_format *f = arg;

		memset(f,0,sizeof(*f));
		f->type = V4L2_BUF_TYPE_CAPTURE;
		f->fmt.pix.width        = fh->width;
		f->fmt.pix.height       = fh->height;
		f->fmt.pix.depth        = fh->fmt->depth;
		f->fmt.pix.pixelformat  = fh->fmt->fourcc;
		f->fmt.pix.sizeimage    =
			(fh->width*fh->height*fh->fmt->depth)/8;
		return 0;
	}
	case VIDIOC_S_FMT:
	{
		struct v4l2_format *f = arg;
		struct saa7134_format *fmt;

		if ((f->type & V4L2_BUF_TYPE_field) != V4L2_BUF_TYPE_CAPTURE)
			return -EINVAL;
		fmt = format_by_fourcc(f->fmt.pix.pixelformat);
		if (NULL == fmt)
			return -EINVAL;

		if (f->fmt.pix.width > dev->tvnorm->width)
			f->fmt.pix.width = dev->tvnorm->width;
		if (f->fmt.pix.height > dev->tvnorm->height)
			f->fmt.pix.height = dev->tvnorm->height;
		
		fh->fmt    = fmt;
		fh->width  = f->fmt.pix.width;
		fh->height = f->fmt.pix.height;
		fh->inter  =
			(f->fmt.pix.flags & V4L2_FMT_FLAG_INTERLACED) &&
			(f->fmt.pix.height > dev->tvnorm->height / 2);

		/* update data for the application */
		f->fmt.pix.depth = fmt->depth;
		f->fmt.pix.sizeimage =
			(fh->width * fh->height * fmt->depth)/8;
		return 0;
	}

	case VIDIOCGMBUF:
	{
		struct video_mbuf *mbuf = arg;
		int i;

		down(&fh->q.lock);
		err = videobuf_mmap_setup(file,&fh->q,gbuffers,gbufsize);
		if (err < 0)
			goto fh_unlock_and_return;
		memset(mbuf,0,sizeof(*mbuf));
		mbuf->frames = gbuffers;
		mbuf->size   = gbuffers * gbufsize;
		for (i = 0; i < gbuffers; i++)
			mbuf->offsets[i] = i * gbufsize;
		up(&fh->q.lock);
		return 0;
	}
	case VIDIOC_REQBUFS:
		return videobuf_reqbufs(file,&fh->q,arg);

	case VIDIOC_QUERYBUF:
		return videobuf_querybuf(&fh->q,arg);

	case VIDIOC_QBUF:
		return videobuf_qbuf(file,&fh->q,arg);

	case VIDIOC_DQBUF:
		return videobuf_dqbuf(file,&fh->q,arg);

	case VIDIOC_STREAMON:
		if (!res_get(dev,fh,RESOURCE_STREAMING))
			return -EBUSY;
		return videobuf_streamon(file,&fh->q);

	case VIDIOC_STREAMOFF:
		err = videobuf_streamoff(file,&fh->q);
		if (err < 0)
			goto fh_unlock_and_return;
		res_free(dev,fh,RESOURCE_STREAMING);
		return 0;
		
	default:
		//up(&fh->q.lock); /* FIXME */
		return v4l_compat_translate_ioctl(inode,file,cmd,arg,
						  video_do_ioctl);
	}
	return 0;

 fh_unlock_and_return:
	up(&fh->q.lock);
	return err;
}

static int video_ioctl(struct inode *inode, struct file *file,
		       unsigned int cmd, unsigned long arg)
{
	return video_usercopy(inode, file, cmd, arg, video_do_ioctl);
}

static int radio_do_ioctl(struct inode *inode, struct file *file,
			  unsigned int cmd, void *arg)
{
	struct saa7134_video_fh *fh = file->private_data;
	struct saa7134_dev *dev = fh->dev;
	int err;
	
	if (video_debug > 1)
		saa7134_print_ioctl(dev->name,cmd);
	switch (cmd) {
	case VIDIOC_QUERYCAP:
	{
		struct v4l2_capability *cap = arg;

		memset(cap,0,sizeof(*cap));
                strcpy(cap->name, dev->name);
		cap->type   = V4L2_TYPE_RADIO;
		cap->flags  = V4L2_FLAG_TUNER;
		cap->inputs = 1;
		cap->audios = 1;
		return 0;
	}
	case VIDIOC_G_TUNER:
	{
		struct v4l2_tuner *t = arg;
		struct video_tuner vt;

		memset(t,0,sizeof(*t));
		strcpy(t->name, "Radio");
                t->rangelow  = (int)(65*16);
                t->rangehigh = (int)(108*16);
		
		memset(&vt,0,sizeof(vt));
		saa7134_i2c_call_clients(dev,VIDIOCGTUNER,&vt);
		t->signal = vt.signal;
		return 0;
	}
	case VIDIOC_ENUMINPUT:
	{
		struct v4l2_input *i = arg;
		
		if (i->index != 0)
			return -EINVAL;
		strcpy(i->name,"Radio");
		i->type = V4L2_INPUT_TYPE_TUNER;
		i->capability = V4L2_INPUT_CAP_AUDIO;
		i->assoc_audio = 0;
		return 0;
	}
	case VIDIOC_G_INPUT:
	{
		int *i = arg;
		*i = 0;
		return 0;
	}
	case VIDIOC_G_AUDIO:
	{
		struct v4l2_audio *a = arg;

		memset(a,0,sizeof(*a));
		strcpy(a->name,"Radio");
		return 0;
	}
	case VIDIOC_S_AUDIO:
	case VIDIOC_S_TUNER:
	case VIDIOC_S_INPUT:
		return 0;

	case VIDIOC_QUERYCTRL:
	{
		const struct v4l2_queryctrl *ctrl;
		struct v4l2_queryctrl *c = arg;

		v4l2_fill_ctrl_category(c);
		if (c->id <  V4L2_CID_BASE ||
		    c->id >= V4L2_CID_LASTP1)
			return -EINVAL;
		if (c->id == V4L2_CID_AUDIO_MUTE) {
			ctrl = ctrl_by_id(c->id);
			*c = *ctrl;
		} else
			*c = no_ctrl;
		return 0;
	}
	case VIDIOC_G_CTRL:
		return get_control(dev,arg);
	case VIDIOC_S_CTRL:
	{
		down(&dev->lock);
		err = set_control(dev,fh,arg);
		up(&dev->lock);
		return err;
	}

	case VIDIOC_G_FREQ:
	case VIDIOC_S_FREQ:
		return saa7134_common_ioctl(dev,cmd,arg);
		
	default:
		//up(&fh->q.lock); /* FIXME */
		return v4l_compat_translate_ioctl(inode,file,cmd,arg,
						  radio_do_ioctl);
	}
	return 0;
}

static int radio_ioctl(struct inode *inode, struct file *file,
		       unsigned int cmd, unsigned long arg)
{
	return video_usercopy(inode, file, cmd, arg, radio_do_ioctl);
}

static struct file_operations video_fops =
{
	owner:	  THIS_MODULE,
	open:	  video_open,
	release:  video_release,
	read:	  video_read,
	poll:     video_poll,
	mmap:	  video_mmap,
	ioctl:	  video_ioctl,
	llseek:   no_llseek,
};

static struct file_operations radio_fops =
{
	owner:	  THIS_MODULE,
	open:	  video_open,
	release:  video_release,
	ioctl:	  radio_ioctl,
	llseek:   no_llseek,
};

/* ----------------------------------------------------------- */
/* exported stuff                                              */

struct video_device saa7134_video_template =
{
	name:          "saa7134-video",
	type:          VID_TYPE_CAPTURE|VID_TYPE_TUNER|VID_TYPE_OVERLAY|
	               VID_TYPE_CLIPPING|VID_TYPE_SCALES,
	type2:         V4L2_TYPE_CAPTURE,
	hardware:      0,
	fops:          &video_fops,
	minor:         -1,
};

struct video_device saa7134_radio_template =
{
	name:          "saa7134-radio",
	type:          VID_TYPE_TUNER,
	type2:         V4L2_TYPE_RADIO,
	hardware:      0,
	fops:          &radio_fops,
	minor:         -1,
};

int saa7134_video_init(struct saa7134_dev *dev)
{
	/* sanitycheck insmod options */
	if (gbuffers < 2 || gbuffers > VIDEO_MAX_FRAME)
		gbuffers = 2;
	if (gbufsize < 0 || gbufsize > gbufsize_max)
		gbufsize = gbufsize_max;
	gbufsize = (gbufsize + PAGE_SIZE - 1) & PAGE_MASK;

	/* put some sensible defaults into the data structures ... */
	dev->ctl_bright     = ctrl_by_id(V4L2_CID_BRIGHTNESS)->default_value;
	dev->ctl_contrast   = ctrl_by_id(V4L2_CID_CONTRAST)->default_value;
	dev->ctl_hue        = ctrl_by_id(V4L2_CID_HUE)->default_value;
	dev->ctl_saturation = ctrl_by_id(V4L2_CID_SATURATION)->default_value;
	dev->ctl_volume     = ctrl_by_id(V4L2_CID_AUDIO_VOLUME)->default_value;

	dev->ctl_invert     = 0;
	dev->ctl_interlace  = INTERLACE_AUTO;
	dev->ctl_mute       = 1;
	dev->automute       = 0;

        INIT_LIST_HEAD(&dev->video_q.queue);
	dev->video_q.timeout.function = saa7134_buffer_timeout;
	dev->video_q.timeout.data     = (unsigned long)(&dev->video_q);
	dev->video_q.dev              = dev;

	/* init video hw */
	set_tvnorm(dev,&tvnorms[0]);
	video_mux(dev,0);
	saa7134_tvaudio_setmute(dev);
	saa7134_tvaudio_setvolume(dev,dev->ctl_volume);

	if (saa7134_boards[dev->board].video_out) {
		/* enable video output */
		int vo = saa7134_boards[dev->board].video_out;
		saa_writeb(SAA7134_VIDEO_PORT_CTRL0, video_out[vo][0]);
		saa_writeb(SAA7134_VIDEO_PORT_CTRL1, video_out[vo][1]);
		saa_writeb(SAA7134_VIDEO_PORT_CTRL2, video_out[vo][2]);
		saa_writeb(SAA7134_VIDEO_PORT_CTRL3, video_out[vo][3]);
		saa_writeb(SAA7134_VIDEO_PORT_CTRL4, video_out[vo][4]);
		saa_writeb(SAA7134_VIDEO_PORT_CTRL5, video_out[vo][5]);
		saa_writeb(SAA7134_VIDEO_PORT_CTRL6, video_out[vo][6]);
		saa_writeb(SAA7134_VIDEO_PORT_CTRL7, video_out[vo][7]);
		saa_writeb(SAA7134_VIDEO_PORT_CTRL8, video_out[vo][8]);
	}
 
	return 0;
}

int saa7134_video_fini(struct saa7134_dev *dev)
{
	/* nothing */
	return 0;
}

void saa7134_irq_video_intl(struct saa7134_dev *dev)
{
	static const char *st[] = {
		"no signal", "found NTSC", "found PAL", "found SECAM" };
	int norm;

	norm = saa_readb(SAA7134_STATUS_VIDEO1) & 0x03;
	printk("%s/video: DCSDT: %s\n",dev->name,st[norm]);
	
	if (0 != norm) {
		/* wake up tvaudio audio carrier scan thread */
		saa7134_tvaudio_do_scan(dev);
	} else {
		/* no video signal -> mute audio */
		dev->automute = 1;
		saa7134_tvaudio_setmute(dev);
	}
}

void saa7134_irq_video_done(struct saa7134_dev *dev, unsigned long status)
{
	spin_lock(&dev->slock);
	if (dev->video_q.curr) {
		if (dev->video_q.curr->vb.field & VBUF_FIELD_INTER) {
			/* make sure we have seen both fields */
			if ((status & 0x10) == 0x10) {
				dev->video_q.curr->odd_seen = 1;
				goto done;
			}
			if (!dev->video_q.curr->odd_seen)
				goto done;
		} else if (dev->video_q.curr->vb.field & VBUF_FIELD_ODD) {
			if ((status & 0x10) != 0x10)
				goto done;
		} else if (dev->video_q.curr->vb.field & VBUF_FIELD_EVEN) {
			if ((status & 0x10) != 0x00)
				goto done;
		}
		saa7134_buffer_finish(dev,&dev->video_q,STATE_DONE);
	}
	saa7134_buffer_next(dev,&dev->video_q);

 done:
	spin_unlock(&dev->slock);
}

/* ----------------------------------------------------------- */
/*
 * Local variables:
 * c-basic-offset: 8
 * End:
 */
